// _readmap.idc
// This script imports MAP file information from the Hitachi Optimizing Linkage Editor
// Ver. 7.0, which comes with the Hitachi Tool Chain 4.0.0, from the Hew 1.2 package.
//
// code by TWH
//
// History:
// 01/26/2000 TWH Initial coding

#include <idc.idc>

// These values for use with Hitachi H8300L (3644F)
#define SEGOFFSET 0x10000
#define RAM_START 0xFB80
#define RAM_END   0xFF7F

// WARNING: this function ignores all unrecognized chars (including lowercase alphas)
//          eg, "1 2-3 A3.4HZb 77" is treated as "123A3477"
static htol(strHex) {
    auto hlt, v, i, j, x;

    hlt = "0123456789ABCDEF";
    v = 0;

    j = strlen(strHex);
    i = 0;
    while (i < j) {
        x = strstr(hlt,  substr(strHex, i, i+1));
	if (x>=0)
	        v = (v << 4) + x;
        i++;
    }

    return(v);
}

static GNFToken(InFileH) {
    auto name, t, whitespace;

    whitespace = "\040\012\015";
    name = "";
    t = fgetc(InFileH);
    while ( ( t != -1 ) && ( strstr(whitespace, t) >= 0 ) ) {
        t = fgetc(InFileH);
    }
    while ( t != -1 ) {
        name = name + t;
        t = fgetc(InFileH);
        if ( strstr(whitespace, t) >= 0 )
            break;
    }

    return(name);
}

static GetInputPath(void) {
    auto p, r, e;

    p = GetInputFilePath();
    r = -1;
    e = strlen( p );
    while ( r < 0 ) {
        e--;
        r = strstr(substr(p, e, -1), "\\");
    }

    return( substr(p, 0, ++e) );
}

// TODO: Actually prompt the user for correct values
static FixIndirectCalls(void) {
    auto ea, lea;
    auto szOpc;
    auto szOpr;
    auto szFN;
    auto szRN;
    auto vea;
    auto vv;
    auto szPrompt;

    ea = FindCode(MinEA(), SEARCH_DOWN);
    Message("Started search for indirect calls @ %s\n", ltoa(ea, 16));

    while (ea != BADADDR) {
        szOpc = GetMnem(ea);					// get instruction text

// Check for "jsr"
        if (szOpc == "jsr") {
            szOpr = GetOpnd(ea, 0);				// get operand text

// Check for "@r"
            if (strstr(szOpr, "@r") == 0) {
                szFN = GetFunctionName(ea);			// get current function name
                szRN = substr(szOpr, 1, 3);			// get current register number

// Determine best guess at the indirect address value
                vea = FindCode(LocByName(GetFunctionName(ea)), SEARCH_DOWN); 
                vv = "0";
                while ((GetFunctionName(vea) == szFN) && (vv == "0")) {
                    if ((GetMnem(vea) == "mov") && (GetOpnd(vea, 1) == szRN)) {
                        vv = GetOpnd(vea, 0);
                        vv = substr(vv, 3, -1);
                    }
                    vea = FindCode(vea, SEARCH_DOWN);
                }

// Prompt user for address value
		szPrompt = form("I think ROM:%s operand @%s is %s(%s)", ltoa(ea - SEGOFFSET, 16), szRN, vv, GetFunctionName(xtol(vv) + SEGOFFSET) );
                vv = AskStr(vv, szPrompt);

// Add the user-defined XREF
                vv = xtol(vv) + SEGOFFSET;
                AddCodeXref(ea, vv, fl_US);            }
        }
	lea = ea;
        ea = FindCode(ea, SEARCH_DOWN);
    }
    Message("Finished search for indirect calls @ %s\n", ltoa(lea, 16));

}

static exists(FileName) {
    auto fh;

    fh = fopen(FileName, "r");
    if (fh) {
        fclose(fh);
        fh = 1;
    }

    return fh;
}

static main(void) {
    auto minea, ea, tkn, name;
    auto InFName, InFileH;
    auto errcode;
    auto i, j;

    minea = MinEA();

// browse for an input symbol file name (.map)
    InFName = AskFile(0, GetInputPath() + "*.map", "Please locate the map file:");
    if (exists(InFName))
    {
        InFileH = fopen(InFName, "rb");
        Message("Fixing up segments...\n");

//      errcode = SegDelete(SEGOFFSET + IVT_START, 0);
//      errcode = SegDelete(SEGOFFSET + ROM_START, 0);
//      errcode = SegDelete(SEGOFFSET + RAM_START, 0);

//      errcode = SegCreate(SEGOFFSET + IVT_START, SEGOFFSET + IVT_END + 1, (SEGOFFSET >> 4), 0, saRelPage, scPub);
//      errcode = SegRename(SEGOFFSET + IVT_START, "IVT");

//      errcode = SegCreate(SEGOFFSET + ROM_START, SEGOFFSET + ROM_START, SEGOFFSET + ROM_END + 1, 0);
//      errcode = SegCreate(SEGOFFSET + ROM_START, SEGOFFSET + ROM_END + 1, (SEGOFFSET >> 4), 0, saRelPage, scPub);
//      errcode = SegRename(SEGOFFSET + ROM_START, "ROM");

        errcode = SegRename(SEGOFFSET + IVT_START, "ROM");

        errcode = SegCreate(SEGOFFSET + RAM_START, SEGOFFSET + RAM_END + 1, (SEGOFFSET >> 4), 0, saRelPage, scPub);
        errcode = SegRename(SEGOFFSET + RAM_START, "RAM");

        Message("Beginning import of Hitachi symbols.\n");

// locate tof marker - BEGIN
        tkn = GNFToken(InFileH);
        while ( tkn != "BEGIN" ) {
            tkn = GNFToken(InFileH);
        }

// watch for eof marker - "END"
        tkn = GNFToken(InFileH);
        while ( tkn != "END" ) {

// calculate address from HEX rep
            ea = htol( tkn );

            ea = ea + minea;
// read the name label
            tkn = GNFToken(InFileH);

            name = tkn;
            if (MakeName( ea, "" ) == 0 ) {
                MakeUnkn(ea, 0);
                MakeName( ea, "" );
            }
            if (MakeName( ea, name ) == 0 ) {
                Message("    Address: '" + tkn + "'" + "(%x) ", ea );
                Message("    Label: '" + name);
                Message("    Failed!\n");
            }
// mark functions
//            MakeFunction( ea, BADADDR );
// mark library functions
//            Message("F: %s, %s\n", name, substr(name, 0, 1));
            if (substr(name, 0, 1) == "$") {
                if (ea < ROM_END) {
                    errcode = SetFunctionFlags( ea, FUNC_LIB );
                }
            }

            tkn = GNFToken(InFileH);
        }

        Message("Finished importing Hitachi symbols.\n");
        fclose (InFileH);

// mark all vector table entries as entry points
        for (i = (SEGOFFSET + IVT_START); i < (SEGOFFSET + ROM_START); i = i + 2 ) {
            MakeUnkn(i, 0);
            MakeUnkn(i+1, 0);
            MakeWord(i);
            OpOff(i,0,SEGOFFSET);
            j = Word( i ) + SEGOFFSET;
            errcode = AddEntryPoint(j,j,"",1);
            MakeCode( j );
            AutoMark( j, AU_PROC );
        }

// add XREFs for indirect calls
        FixIndirectCalls();

// get Power On Reset vector
        i = Word(SEGOFFSET + IVT_START) + SEGOFFSET;
        Message("Entry point is: %08x\n", i );
        AnalyseArea(i, SEGOFFSET + ROM_END + 1);
        Wait();

    }

}

